/*
 * snul.vala is a part of Snul
 *
 * Snul: is a simple library that implements
 *       a tcp socket client in vala
 *
 * Copyright (C) 2008  Andrea Del Signore
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 *
 * Author:
 *     Andrea Del Signore <sejerpz@tin.it>
 */

#include <src/snul.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>

struct _SnulTcpClientSocketPrivate {
	char* _address;
	char* _first_address;
	char* _port;
	gushort _port_number;
	gint _socket_fd;
	GIOChannel* _io_channel;
	guint _watch_id;
};
#define SNUL_TCP_CLIENT_SOCKET_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SNUL_TYPE_TCP_CLIENT_SOCKET, SnulTcpClientSocketPrivate))
enum  {
	SNUL_TCP_CLIENT_SOCKET_DUMMY_PROPERTY,
	SNUL_TCP_CLIENT_SOCKET_ADDRESS,
	SNUL_TCP_CLIENT_SOCKET_PORT
};
static const char* SNUL_TCP_CLIENT_SOCKET_error_domain_quark = "tcp_socket_error_domain-quark";
static void snul_tcp_client_socket_on_connected_real (SnulTcpClientSocket* self, GError** error);
static gboolean snul_tcp_client_socket_socket_callback (GIOChannel* channel, GIOCondition condition, SnulTcpClientSocket* self);
static gpointer snul_tcp_client_socket_parent_class = NULL;
static void snul_tcp_client_socket_dispose (GObject * obj);

static void g_cclosure_user_marshal_VOID__STRING_ULONG (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data);

GQuark snul_tcp_socket_error_quark (void) {
	return g_quark_from_static_string ("snul_tcp_socket_error-quark");
}


gint snul_tcp_client_socket_connect (SnulTcpClientSocket* self, const char* address, const char* port, GError** error) {
	GError * inner_error;
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self), 0);
	inner_error = NULL;
	{
		char* _tmp1;
		const char* _tmp0;
		char* _tmp3;
		const char* _tmp2;
		struct hostent* hp;
		char* _tmp5;
		const char* _tmp4;
		struct sockaddr_in* sin;
		struct in_addr* add;
		gint _tmp6;
		_tmp1 = NULL;
		_tmp0 = NULL;
		self->priv->_address = (_tmp1 = (_tmp0 = address, (_tmp0 == NULL ? NULL : g_strdup (_tmp0))), (self->priv->_address = (g_free (self->priv->_address), NULL)), _tmp1);
		_tmp3 = NULL;
		_tmp2 = NULL;
		self->priv->_port = (_tmp3 = (_tmp2 = port, (_tmp2 == NULL ? NULL : g_strdup (_tmp2))), (self->priv->_port = (g_free (self->priv->_port), NULL)), _tmp3);
		if (self->priv->_socket_fd > 0) {
			g_set_error (error, SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_ALREADY_OPEN, "Socket already opened");
			return 0;
		}
		if (g_unichar_isdigit (g_utf8_get_char (g_utf8_offset_to_pointer (self->priv->_port, 0)))) {
			self->priv->_port_number = ((gushort) atoi (self->priv->_port));
		} else {
			struct servent* sp;
			sp = getservbyname (self->priv->_port, "tcp");
			if (sp == NULL) {
				g_set_error (error, SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_LOOKUP_SERVICE_FAILED, "Lookup service by name failed");
				return 0;
			}
			self->priv->_port_number = ((gushort) sp->s_port);
		}
		hp = gethostbyname (self->priv->_address);
		if (hp == NULL) {
			g_set_error (error, SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_LOOKUP_HOSTNAME_FAILED, "Lookup host by name failed");
			return 0;
		}
		_tmp5 = NULL;
		_tmp4 = NULL;
		self->priv->_first_address = (_tmp5 = (_tmp4 = hp->h_addr_list[0], (_tmp4 == NULL ? NULL : g_strdup (_tmp4))), (self->priv->_first_address = (g_free (self->priv->_first_address), NULL)), _tmp5);
		self->priv->_socket_fd = socket (AF_INET, SOCK_STREAM, 0);
		if (self->priv->_socket_fd < 0) {
			g_set_error (error, SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_CREATION_ERROR, "Socket creation failed");
			return 0;
		}
		sin = g_new0 (struct sockaddr_in, 1);
		add = hp->h_addr_list[0];
		sin->sin_family = AF_INET;
		sin->sin_addr.s_addr = add->s_addr;
		sin->sin_port = htons (self->priv->_port_number);
		if (connect (self->priv->_socket_fd, ((struct sockaddr*) sin), ((guint) sizeof (struct sockaddr))) < 0) {
			g_set_error (error, SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_CONNECT_ERROR, "Connect error");
			return 0;
		}
		snul_tcp_client_socket_on_connected_real (self, &inner_error);
		if (inner_error != NULL) {
			if (inner_error->domain == SNUL_TCP_SOCKET_ERROR) {
				goto __catch1_snul_tcp_socket_error;
			}
			g_critical ("file %s: line %d: uncaught error: %s", __FILE__, __LINE__, inner_error->message);
			g_clear_error (&inner_error);
		}
		return (_tmp6 = self->priv->_socket_fd, (sin == NULL ? NULL : (sin = (free (sin), NULL))), _tmp6);
	}
	goto __finally1;
	__catch1_snul_tcp_socket_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			snul_tcp_client_socket_on_error (self, err);
			err;
			return 0;
		}
	}
	__finally1:
	;
}


gsize snul_tcp_client_socket_send (SnulTcpClientSocket* self, const char* buffer) {
	GError * inner_error;
	gssize len;
	gsize bytes_written;
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self), 0UL);
	inner_error = NULL;
	if (buffer == NULL) {
		return 0;
	}
	len = strlen (buffer);
	bytes_written = 0UL;
	{
		g_io_channel_write_chars (self->priv->_io_channel, ((gchar*) buffer), ((gssize) len), &bytes_written, &inner_error);
		if (inner_error != NULL) {
			goto __catch2_g_error;
		}
		g_io_channel_flush (self->priv->_io_channel, &inner_error);
		if (inner_error != NULL) {
			goto __catch2_g_error;
		}
	}
	goto __finally2;
	__catch2_g_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			g_warning ("snul.vala:145: error writing: %s", err->message);
		}
	}
	__finally2:
	;
	return bytes_written;
}


void snul_tcp_client_socket_disconnect (SnulTcpClientSocket* self) {
	GError * inner_error;
	g_return_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self));
	inner_error = NULL;
	if (self->priv->_watch_id != 0) {
		g_source_remove (self->priv->_watch_id);
		self->priv->_watch_id = 0;
	}
	if (self->priv->_io_channel != NULL) {
		GIOChannel* _tmp0;
		{
			g_io_channel_shutdown (self->priv->_io_channel, TRUE, &inner_error);
			if (inner_error != NULL) {
				goto __catch3_g_error;
			}
		}
		goto __finally3;
		__catch3_g_error:
		{
			GError * err;
			err = inner_error;
			inner_error = NULL;
			{
				g_warning ("snul.vala:161: error shutting down channel: %s", err->message);
			}
		}
		__finally3:
		;
		_tmp0 = NULL;
		self->priv->_io_channel = (_tmp0 = NULL, (self->priv->_io_channel == NULL ? NULL : (self->priv->_io_channel = (g_io_channel_unref (self->priv->_io_channel), NULL))), _tmp0);
	}
	if (self->priv->_socket_fd != 0) {
		close (self->priv->_socket_fd);
	}
	snul_tcp_client_socket_on_disconnected (self);
}


void snul_tcp_client_socket_on_connected (SnulTcpClientSocket* self, gint socket_fd) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self));
	g_signal_emit_by_name (G_OBJECT (self), "connected", socket_fd);
}


void snul_tcp_client_socket_on_data_received (SnulTcpClientSocket* self, const char* buffer, gsize len) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self));
	g_signal_emit_by_name (G_OBJECT (self), "data-received", buffer, len);
}


void snul_tcp_client_socket_on_disconnected (SnulTcpClientSocket* self) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self));
	g_signal_emit_by_name (G_OBJECT (self), "disconnected");
}


void snul_tcp_client_socket_on_error (SnulTcpClientSocket* self, GError* error) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self));
	g_signal_emit_by_name (G_OBJECT (self), "error", error);
}


static void snul_tcp_client_socket_on_connected_real (SnulTcpClientSocket* self, GError** error) {
	GError * inner_error;
	GIOChannel* _tmp0;
	GIOCondition conditions;
	g_return_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self));
	inner_error = NULL;
	if (self->priv->_socket_fd <= 0) {
		g_set_error (error, SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_CLOSED, "Closed or bad socket");
		return;
	}
	_tmp0 = NULL;
	self->priv->_io_channel = (_tmp0 = g_io_channel_unix_new (self->priv->_socket_fd), (self->priv->_io_channel == NULL ? NULL : (self->priv->_io_channel = (g_io_channel_unref (self->priv->_io_channel), NULL))), _tmp0);
	{
		g_io_channel_set_encoding (self->priv->_io_channel, NULL, &inner_error);
		if (inner_error != NULL) {
			if (inner_error->domain == G_CONVERT_ERROR) {
				goto __catch4_g_convert_error;
			}
			g_critical ("file %s: line %d: uncaught error: %s", __FILE__, __LINE__, inner_error->message);
			g_clear_error (&inner_error);
		}
	}
	goto __finally4;
	__catch4_g_convert_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			err;
			return;
		}
	}
	__finally4:
	;
	conditions = G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL;
	self->priv->_watch_id = g_io_add_watch (self->priv->_io_channel, conditions, ((UtilsIOFunc) snul_tcp_client_socket_socket_callback), self);
	snul_tcp_client_socket_on_connected (self, self->priv->_socket_fd);
}


static gboolean snul_tcp_client_socket_socket_callback (GIOChannel* channel, GIOCondition condition, SnulTcpClientSocket* self) {
	GError * inner_error;
	gboolean res;
	GIOCondition _tmp6;
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self), FALSE);
	inner_error = NULL;
	res = TRUE;
	_tmp6 = condition;
	if (_tmp6 == G_IO_IN)
	do {
		char* buffer;
		gsize len;
		gsize term_pos;
		buffer = NULL;
		len = 0UL;
		term_pos = 0UL;
		{
			GIOStatus _tmp1;
			char* _tmp0;
			_tmp0 = NULL;
			(_tmp1 = g_io_channel_read_line (self->priv->_io_channel, &_tmp0, &len, &term_pos, &inner_error), (buffer = (g_free (buffer), NULL)), buffer = _tmp0, _tmp1);
			if (inner_error != NULL) {
				goto __catch5_g_error;
			}
			if (len > 0) {
				snul_tcp_client_socket_on_data_received (self, buffer, len);
			} else {
				GError* _tmp2;
				_tmp2 = NULL;
				snul_tcp_client_socket_on_error (self, (_tmp2 = g_error_new (g_quark_from_string (SNUL_TCP_CLIENT_SOCKET_error_domain_quark), SNUL_TCP_SOCKET_ERROR_SOCKET_BROKEN, "zero length data, socket broken? Closing connection.", NULL)));
				(_tmp2 == NULL ? NULL : (_tmp2 = (g_error_free (_tmp2), NULL)));
				snul_tcp_client_socket_disconnect (self);
			}
		}
		goto __finally5;
		__catch5_g_error:
		{
			GError * err;
			err = inner_error;
			inner_error = NULL;
			{
				g_warning ("snul.vala:230: error reading: %s", err->message);
			}
		}
		__finally5:
		;
		(buffer = (g_free (buffer), NULL));
		break;
	} while (0); else if (_tmp6 == G_IO_HUP)
	do {
		GError* _tmp3;
		_tmp3 = NULL;
		snul_tcp_client_socket_on_error (self, (_tmp3 = g_error_new (g_quark_from_string (SNUL_TCP_CLIENT_SOCKET_error_domain_quark), SNUL_TCP_SOCKET_ERROR_SOCKET_HUP, "IOChannel error", NULL)));
		(_tmp3 == NULL ? NULL : (_tmp3 = (g_error_free (_tmp3), NULL)));
		snul_tcp_client_socket_disconnect (self);
		res = FALSE;
		break;
	} while (0); else if (_tmp6 == G_IO_ERR)
	do {
		GError* _tmp4;
		_tmp4 = NULL;
		snul_tcp_client_socket_on_error (self, (_tmp4 = g_error_new (g_quark_from_string (SNUL_TCP_CLIENT_SOCKET_error_domain_quark), SNUL_TCP_SOCKET_ERROR_IOCHANNEL_ERROR, "IOChannel error", NULL)));
		(_tmp4 == NULL ? NULL : (_tmp4 = (g_error_free (_tmp4), NULL)));
		snul_tcp_client_socket_disconnect (self);
		break;
	} while (0); else if (_tmp6 == G_IO_NVAL)
	do {
		GError* _tmp5;
		_tmp5 = NULL;
		snul_tcp_client_socket_on_error (self, (_tmp5 = g_error_new (g_quark_from_string (SNUL_TCP_CLIENT_SOCKET_error_domain_quark), SNUL_TCP_SOCKET_ERROR_IOCHANNEL_NVAL, "IOChannel invalid request", NULL)));
		(_tmp5 == NULL ? NULL : (_tmp5 = (g_error_free (_tmp5), NULL)));
		snul_tcp_client_socket_disconnect (self);
		break;
	} while (0);
	return res;
}


SnulTcpClientSocket* snul_tcp_client_socket_new (void) {
	SnulTcpClientSocket * self;
	self = g_object_newv (SNUL_TYPE_TCP_CLIENT_SOCKET, 0, NULL);
	return self;
}


char* snul_tcp_client_socket_get_address (SnulTcpClientSocket* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self), NULL);
	return self->priv->_address;
}


char* snul_tcp_client_socket_get_port (SnulTcpClientSocket* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT_SOCKET (self), NULL);
	return self->priv->_port;
}


static void snul_tcp_client_socket_get_property (GObject * object, guint property_id, GValue * value, GParamSpec * pspec) {
	SnulTcpClientSocket * self;
	self = SNUL_TCP_CLIENT_SOCKET (object);
	switch (property_id) {
		case SNUL_TCP_CLIENT_SOCKET_ADDRESS:
		g_value_set_string (value, snul_tcp_client_socket_get_address (self));
		break;
		case SNUL_TCP_CLIENT_SOCKET_PORT:
		g_value_set_string (value, snul_tcp_client_socket_get_port (self));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void snul_tcp_client_socket_class_init (SnulTcpClientSocketClass * klass) {
	snul_tcp_client_socket_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (SnulTcpClientSocketPrivate));
	G_OBJECT_CLASS (klass)->get_property = snul_tcp_client_socket_get_property;
	G_OBJECT_CLASS (klass)->dispose = snul_tcp_client_socket_dispose;
	g_object_class_install_property (G_OBJECT_CLASS (klass), SNUL_TCP_CLIENT_SOCKET_ADDRESS, g_param_spec_string ("address", "address", "address", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), SNUL_TCP_CLIENT_SOCKET_PORT, g_param_spec_string ("port", "port", "port", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_signal_new ("connected", SNUL_TYPE_TCP_CLIENT_SOCKET, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);
	g_signal_new ("data_received", SNUL_TYPE_TCP_CLIENT_SOCKET, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_user_marshal_VOID__STRING_ULONG, G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_ULONG);
	g_signal_new ("disconnected", SNUL_TYPE_TCP_CLIENT_SOCKET, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
	g_signal_new ("error", SNUL_TYPE_TCP_CLIENT_SOCKET, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__POINTER, G_TYPE_NONE, 1, G_TYPE_POINTER);
}


static void snul_tcp_client_socket_init (SnulTcpClientSocket * self) {
	self->priv = SNUL_TCP_CLIENT_SOCKET_GET_PRIVATE (self);
	self->priv->_address = NULL;
	self->priv->_first_address = NULL;
	self->priv->_port = NULL;
	self->priv->_port_number = 0;
	self->priv->_socket_fd = 0;
	self->priv->_io_channel = NULL;
	self->priv->_watch_id = 0;
}


static void snul_tcp_client_socket_dispose (GObject * obj) {
	SnulTcpClientSocket * self;
	self = SNUL_TCP_CLIENT_SOCKET (obj);
	(self->priv->_address = (g_free (self->priv->_address), NULL));
	(self->priv->_first_address = (g_free (self->priv->_first_address), NULL));
	(self->priv->_port = (g_free (self->priv->_port), NULL));
	(self->priv->_io_channel == NULL ? NULL : (self->priv->_io_channel = (g_io_channel_unref (self->priv->_io_channel), NULL)));
	G_OBJECT_CLASS (snul_tcp_client_socket_parent_class)->dispose (obj);
}


GType snul_tcp_client_socket_get_type (void) {
	static GType snul_tcp_client_socket_type_id = 0;
	if (G_UNLIKELY (snul_tcp_client_socket_type_id == 0)) {
		static const GTypeInfo g_define_type_info = { sizeof (SnulTcpClientSocketClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) snul_tcp_client_socket_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (SnulTcpClientSocket), 0, (GInstanceInitFunc) snul_tcp_client_socket_init };
		snul_tcp_client_socket_type_id = g_type_register_static (G_TYPE_OBJECT, "SnulTcpClientSocket", &g_define_type_info, 0);
	}
	return snul_tcp_client_socket_type_id;
}



static void g_cclosure_user_marshal_VOID__STRING_ULONG (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data) {
	typedef void (*GMarshalFunc_VOID__STRING_ULONG) (gpointer data1, const char* arg_1, gsize arg_2, gpointer data2);
	register GMarshalFunc_VOID__STRING_ULONG callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = ((GCClosure *) closure);
	g_return_if_fail (n_param_values == 3);
	if (G_CCLOSURE_SWAP_DATA (closure)) {
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else {
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = ((GMarshalFunc_VOID__STRING_ULONG) (marshal_data ? marshal_data : cc->callback));
	callback (data1, g_value_get_string (param_values + 1), g_value_get_ulong (param_values + 2), data2);
}



