/*
 * snul.vala is a part of Snul
 *
 * Snul: is a simple library that implements
 *       a tcp socket client in vala
 *
 * Copyright (C) 2008  Andrea Del Signore
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 *
 * Author:
 *     Andrea Del Signore <sejerpz@tin.it>
 */

using GLib;
using UnixSocket;

namespace Utils
{
	[Import]
	[CCode (cname="g_io_channel_set_encoding")]
        public static GLib.IOStatus io_channel_set_encoding (IOChannel channel, string? encoding) throws ConvertError;

	public static delegate bool IOFunc (IOChannel source, IOCondition conditions, pointer data);

	[Import]
	[CCode (cname="g_io_add_watch")]
	public static uint io_add_watch (IOChannel channel, IOCondition conditions,IOFunc func, pointer data);

	[Import]
	[CCode (cname="close")]
	public static void close (int fd);


}

namespace Snul
{
	[ErrorDomain]
	public enum TcpSocketError
	{
                LOOKUP_SERVICE_FAILED,
                LOOKUP_HOSTNAME_FAILED,
                SOCKET_CREATION_ERROR,
                SOCKET_CONNECT_ERROR,
                SOCKET_ALREADY_OPEN,
		SOCKET_CLOSED,
		SOCKET_BROKEN,
		SOCKET_HUP,
		IOCHANNEL_ERROR,
		IOCHANNEL_NVAL
	}

	public class TcpClientSocket : GLib.Object
	{
		public signal void connected (int socket_fd);
		public signal void data_received (string buffer, size_t len);
		public signal void disconnected ();
		public signal void error (Error error);

		private const string error_domain_quark = "tcp_socket_error_domain-quark";

		private string _address = null;
		private string _first_address = null;
		private string _port = null;
		private ushort _port_number = 0;
		private int _socket_fd = 0;

		private IOChannel _io_channel = null;
		private uint _watch_id = 0;

		public string address { get { return _address; } }
		public string port { get { return _port; } }

		public int connect (string address, string port) throws TcpSocketError
		{
			try {
				this._address = address;
				this._port = port;

				if (_socket_fd > 0)
					throw new TcpSocketError.SOCKET_ALREADY_OPEN ("Socket already opened");

				if (_port[0].isdigit()) {
					_port_number = (ushort) _port.to_int();
				} else {
					weak servent sp = getservbyname (_port, "tcp");
					if (sp == null) {
						throw new TcpSocketError.LOOKUP_SERVICE_FAILED ("Lookup service by name failed");
					}
					_port_number = (ushort) sp.s_port;
				}

				weak hostent hp = gethostbyname (_address);
				if (hp == null) {
					throw new TcpSocketError.LOOKUP_HOSTNAME_FAILED ("Lookup host by name failed");
				}
				_first_address = hp.h_addr_list[0];

				_socket_fd = socket (AddressFamilies.INET, SocketTypes.STREAM, 0);
				if (_socket_fd < 0) {
					throw new TcpSocketError.SOCKET_CREATION_ERROR ("Socket creation failed");
				}

				sockaddr_in sin = new sockaddr_in ();
				in_addr *add = hp.h_addr_list[0];
				sin.sin_family = AddressFamilies.INET;
				sin.sin_addr.s_addr = add->s_addr;
				sin.sin_port = UnixSocket.htons(_port_number);
				if (UnixSocket.connect ( _socket_fd, (sockaddr) sin, (uint) sizeof (sockaddr)) < 0) {
					throw new TcpSocketError.SOCKET_CONNECT_ERROR ("Connect error");
				}
				on_connected_real ();
				return _socket_fd;
			} catch (TcpSocketError err) {
				on_error (err);
				throw err;
			}
		}

		public size_t send (string buffer)
		{
			if (buffer == null)
				return 0;

			ssize_t len = buffer.size ();
			size_t bytes_written;

			try {
				_io_channel.write_chars ((char[])buffer, (ssize_t) len, out bytes_written);
				_io_channel.flush ();
			} catch (Error err) {
				warning ("error writing: %s", err.message);
			}
			return bytes_written;
		}

		public void disconnect ()
		{
			if (_watch_id != 0) {
				Source.remove (_watch_id);
				_watch_id = 0;
			}

			if (_io_channel != null) {
				try {
					_io_channel.shutdown (true);
				} catch (Error err) {
					warning ("error shutting down channel: %s", err.message);
				}
				_io_channel = null;
			}

			if (_socket_fd != 0)
                                Utils.close (_socket_fd);

			on_disconnected ();
		}

		protected void on_connected (int socket_fd)
		{
			connected (socket_fd);
		}

		protected void on_data_received (string buffer, size_t len)
		{
			data_received (buffer, len);
		}

		protected void on_disconnected ()
		{
			disconnected ();
		}

		protected void on_error (Error error)
		{
			this.error (error);
		}

		private void on_connected_real () throws TcpSocketError, ConvertError
		{
			if (_socket_fd <= 0)
				throw new TcpSocketError.SOCKET_CLOSED ("Closed or bad socket");

			_io_channel = new IOChannel.unix_new (_socket_fd);
			try {
				Utils.io_channel_set_encoding (_io_channel, null);
			} catch (ConvertError err) {
				throw err;
			}

			IOCondition conditions = IOCondition.IN | IOCondition.HUP | IOCondition.ERR | IOCondition.NVAL;
			_watch_id = Utils.io_add_watch (_io_channel, conditions, this.socket_callback, this);

			on_connected (_socket_fd);
		}

		[InstanceLast]
		private bool socket_callback (IOChannel channel, IOCondition condition)
		{
			bool res = true;
			switch (condition) {
				case IOCondition.IN:
					string buffer;
					size_t len, term_pos;

					try {
						_io_channel.read_line (out buffer, out len, out term_pos);
						if (len > 0)
							on_data_received (buffer, len);
						else {
							on_error (new Error (Quark.from_string (error_domain_quark), 
								TcpSocketError.SOCKET_BROKEN, 
								"zero length data, socket broken? Closing connection."));
							this.disconnect ();
						}
					} catch (Error err) {
						warning ("error reading: %s", err.message);
					}
					break;
				case IOCondition.HUP:
					on_error (new Error (Quark.from_string (error_domain_quark), TcpSocketError.SOCKET_HUP, "IOChannel error"));
					this.disconnect ();
					res = false;
					break;
				case IOCondition.ERR:
					on_error (new Error (Quark.from_string (error_domain_quark), TcpSocketError.IOCHANNEL_ERROR, "IOChannel error"));
					this.disconnect ();
					break;
				case IOCondition.NVAL:
					on_error (new Error (Quark.from_string (error_domain_quark), TcpSocketError.IOCHANNEL_NVAL , "IOChannel invalid request"));
					this.disconnect ();
					break;
			}
			return res;
		}

	}
}